﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2017 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Linq;
using System.Threading;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.Parameters;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Problems.DataAnalysis;

namespace HeuristicLab.Algorithms.DataAnalysis {
  [StorableClass]
  [Item("LogisticLeaf", "A leaf type that uses linear models with a logistic dampening as leaf models. Dampening reduces prediction values far outside the observed target values.")]
  public class LogisticLeaf : ParameterizedNamedItem, ILeafModel {
    private const string DampeningParameterName = "Dampening";
    public IFixedValueParameter<DoubleValue> DampeningParameter {
      get { return Parameters[DampeningParameterName] as IFixedValueParameter<DoubleValue>; }
    }
    public double Dampening {
      get { return DampeningParameter.Value.Value; }
    }

    #region Constructors & Cloning
    [StorableConstructor]
    private LogisticLeaf(bool deserializing) : base(deserializing) { }
    private LogisticLeaf(LogisticLeaf original, Cloner cloner) : base(original, cloner) { }
    public LogisticLeaf() {
      Parameters.Add(new FixedValueParameter<DoubleValue>(DampeningParameterName, "Determines the strenght of the logistic dampening. Must be  > 0.0. Larger numbers make more conservative predictions.", new DoubleValue(1.5)));
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new LogisticLeaf(this, cloner);
    }
    #endregion

    #region IModelType
    public bool ProvidesConfidence {
      get { return true; }
    }
    public IRegressionModel Build(IRegressionProblemData pd, IRandom random, CancellationToken cancellationToken, out int noParameters) {
      if (pd.Dataset.Rows < MinLeafSize(pd)) throw new ArgumentException("The number of training instances is too small to create a linear model");
      double rmse, cvRmse;
      noParameters = pd.AllowedInputVariables.Count() + 1;
      return new DampenedLinearModel(PreconstructedLinearModel.CreateConfidenceLinearModel(pd, out rmse, out cvRmse), pd, Dampening);
    }

    public int MinLeafSize(IRegressionProblemData pd) {
      return pd.AllowedInputVariables.Count() + 2;
    }
    #endregion
  }
}